import os
import unittest
try:
    from packaging.version import Version
except ImportError:
    try:
        from distutils.version import StrictVersion as Version
    except ImportError:
        from setuptools._distutils.version import StrictVersion as Version

from walrus import Database


HOST = os.environ.get('WALRUS_REDIS_HOST') or '127.0.0.1'
PORT = os.environ.get('WALRUS_REDIS_PORT') or 6379

db = Database(host=HOST, port=PORT, db=15)


REDIS_VERSION = None


def requires_version(min_version):
    def decorator(fn):
        global REDIS_VERSION
        if REDIS_VERSION is None:
            REDIS_VERSION = db.info()['redis_version']
        too_old = Version(REDIS_VERSION) < Version(min_version)
        return unittest.skipIf(too_old,
                               'redis too old, requires %s' % min_version)(fn)
    return decorator


def stream_test(fn):
    test_stream = os.environ.get('TEST_STREAM')
    if not test_stream:
        return requires_version('4.9.101')(fn)
    else:
        return unittest.skipIf(not test_stream, 'skipping stream tests')(fn)


def zpop_test(fn):
    test_zpop = os.environ.get('TEST_ZPOP')
    if not test_zpop:
        return requires_version('4.9.101')(fn)
    else:
        return unittest.skipIf(not test_zpop, 'skipping zpop* tests')(fn)


class WalrusTestCase(unittest.TestCase):
    def setUp(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def tearDown(self):
        db.flushdb()
        db._transaction_local.pipes = []

    def assertList(self, values, expected):
        values = list(values)
        self.assertEqual(len(values), len(expected))
        for value, item in zip(values, expected):
            self.assertEqual(value, item)



import unittest

from walrus.containers import *
from walrus.containers import _normalize_stream_keys
from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db
from walrus.tests.base import stream_test
from walrus.tests.base import zpop_test
from walrus.utils import decode
from walrus.utils import decode_dict
from walrus.utils import encode


class TestHash(WalrusTestCase):
    def setUp(self):
        super(TestHash, self).setUp()
        self.hsh = db.Hash('my-hash')

    def test_item_api(self):
        self.hsh['k1'] = 'v1'
        self.assertEqual(self.hsh['k1'], b'v1')
        self.assertTrue(self.hsh['kx'] is None)

        self.hsh['k2'] = 'v2'
        self.hsh['k3'] = 'v3'
        self.assertEqual(self.hsh.as_dict(), {
            b'k1': b'v1',
            b'k2': b'v2',
            b'k3': b'v3'})

        del self.hsh['k2']
        self.assertEqual(self.hsh.as_dict(), {b'k1': b'v1', b'k3': b'v3'})

    def test_dict_apis(self):
        self.hsh.update({'k1': 'v1', 'k2': 'v2'})
        self.hsh.update(k3='v3', k4='v4')
        self.assertEqual(sorted(self.hsh.items()), [
            (b'k1', b'v1'),
            (b'k2', b'v2'),
            (b'k3', b'v3'),
            (b'k4', b'v4')])
        self.assertEqual(sorted(self.hsh.keys()), [b'k1', b'k2', b'k3', b'k4'])
        self.assertEqual(sorted(self.hsh.values()),
                         [b'v1', b'v2', b'v3', b'v4'])

        self.assertEqual(len(self.hsh), 4)
        self.assertTrue('k1' in self.hsh)
        self.assertFalse('kx' in self.hsh)

    def test_search_iter(self):
        self.hsh.update(foo='v1', bar='v2', baz='v3')
        self.assertEqual(sorted(self.hsh), [
            (b'bar', b'v2'),
            (b'baz', b'v3'),
            (b'foo', b'v1')])
        self.assertEqual(sorted(self.hsh.search('b*')), [
            (b'bar', b'v2'),
            (b'baz', b'v3')])

    def test_as_dict(self):
        self.hsh.update(k1='v1', k2='v2')
        self.assertEqual(self.hsh.as_dict(True), {'k1': 'v1', 'k2': 'v2'})
        self.assertEqual(db.Hash('test').as_dict(), {})

    def test_from_dict(self):
        data = dict(zip('abcdefghij', 'klmnopqrst'))
        hsh = Hash.from_dict(db, 'test', data)
        self.assertEqual(hsh.as_dict(True), data)

    def test_setnx(self):
        key, value = "key_setnx", "value_setnx"
        self.assertTrue(self.hsh.setnx(key, value))
        self.assertFalse(self.hsh.setnx(key, value))


class TestSet(WalrusTestCase):
    def setUp(self):
        super(TestSet, self).setUp()
        self.set = db.Set('my-set')

    def test_basic_apis(self):
        self.set.add('i1', 'i2', 'i3', 'i2', 'i1')
        self.assertEqual(sorted(self.set), [b'i1', b'i2', b'i3'])

        self.set.remove('i2')
        self.assertEqual(sorted(self.set), [b'i1', b'i3'])

        self.set.remove('ix')
        self.assertEqual(sorted(self.set), [b'i1', b'i3'])

        # Test __contains__
        self.assertTrue('i1' in self.set)
        self.assertFalse('ix' in self.set)

        # Test __iter__.
        self.assertEqual(sorted(self.set), [b'i1', b'i3'])

        del self.set['i3']
        self.assertEqual(sorted(self.set), [b'i1'])

    def test_combining(self):
        self.set2 = db.Set('my-set2')
        self.set.add(1, 2, 3, 4)
        self.set2.add(3, 4, 5, 6)

        self.assertEqual(
            self.set | self.set2,
            set([b'1', b'2', b'3', b'4', b'5', b'6']))
        self.assertEqual(self.set & self.set2, set([b'3', b'4']))
        self.assertEqual(self.set - self.set2, set([b'1', b'2']))
        self.assertEqual(self.set2 - self.set, set([b'5', b'6']))

    def test_combine_store(self):
        self.set2 = db.Set('my-set2')
        self.set.add(1, 2, 3, 4)
        self.set2.add(3, 4, 5, 6)

        s3 = self.set.unionstore('my-set3', self.set2)
        self.assertEqual(s3.members(),
                         set([b'1', b'2', b'3', b'4', b'5', b'6']))

        s3 = self.set.interstore('my-set3', self.set2)
        self.assertEqual(s3.members(), set([b'3', b'4']))

        s3 = self.set.diffstore('my-set3', self.set2)
        self.assertEqual(s3.members(), set([b'1', b'2']))

        self.set |= self.set2
        self.assertEqual(sorted(self.set),
                         [b'1', b'2', b'3', b'4', b'5', b'6'])

        s4 = db.Set('my-set4')
        s4.add('1', '3')
        s3 &= s4
        self.assertEqual(s3.members(), set([b'1']))

    def test_search(self):
        self.set.add('foo', 'bar', 'baz', 'nug')
        self.assertEqual(sorted(self.set.search('b*')), [b'bar', b'baz'])

    def test_sort(self):
        values = ['charlie', 'zaizee', 'mickey', 'huey']
        self.set.add(*values)
        self.assertEqual(self.set.sort(),
                         [b'charlie', b'huey', b'mickey', b'zaizee'])

        self.set.sort(ordering='DESC', limit=3, store='s_dest')
        self.assertList(db.List('s_dest'), [b'zaizee', b'mickey', b'huey'])

    def test_as_set(self):
        self.set.add('foo', 'bar', 'baz')
        self.assertEqual(self.set.as_set(True), set(('foo', 'bar', 'baz')))
        self.assertEqual(db.Set('test').as_set(), set())

    def test_from_set(self):
        data = set('abcdefghij')
        s = Set.from_set(db, 'test', data)
        self.assertEqual(s.as_set(True), data)


class TestZSet(WalrusTestCase):
    def setUp(self):
        super(TestZSet, self).setUp()
        self.zs = db.ZSet('my-zset')

    def assertZSet(self, expected):
        self.assertEqual(list(self.zs), expected)

    def test_basic_apis(self):
        self.zs.add({'i1': 1, 'i2': 2})
        self.assertZSet([(b'i1', 1), (b'i2', 2)])

        self.zs.add({'i0': 0})
        self.zs.add({'i3': 3})
        self.assertZSet([(b'i0', 0), (b'i1', 1), (b'i2', 2), (b'i3', 3)])

        self.zs.remove('i1')
        self.zs.remove_by_score(3)
        self.zs.add({'i2': -2})
        self.zs.add({'i9': 9})
        self.assertZSet([(b'i2', -2.), (b'i0', 0.), (b'i9', 9.)])

        # __len__
        self.assertEqual(len(self.zs), 3)

        # __contains__
        self.assertTrue('i0' in self.zs)
        self.assertFalse('i1' in self.zs)

        self.assertEqual(self.zs.score('i2'), -2)
        self.assertEqual(self.zs.score('ix'), None)

        self.assertEqual(self.zs.rank('i0'), 1)
        self.assertEqual(self.zs.rank('i1'), None)

        self.assertEqual(self.zs.count(0, 10), 2)
        self.assertEqual(self.zs.count(-3, 11), 3)

        self.zs.incr('i2')
        self.zs.incr('i0', -2)
        self.assertZSet([(b'i0', -2.), (b'i2', -1.), (b'i9', 9.)])

        self.assertEqual(self.zs.range_by_score(0, 9), [b'i9'])
        self.assertEqual(self.zs.range_by_score(-3, 0), [b'i0', b'i2'])

        self.assertEqual(self.zs.popmin_compat(), [(b'i0', -2.)])
        self.assertEqual(len(self.zs), 2)
        self.assertEqual(self.zs.popmax_compat(3),
                              [(b'i9', 9.), (b'i2', -1.)])
        self.assertEqual(self.zs.popmin_compat(), [])
        self.assertEqual(self.zs.popmax_compat(), [])
        self.assertEqual(len(self.zs), 0)

    @zpop_test
    def test_popmin_popmax(self):
        for i in range(10):
            self.zs.add({'i%s' % i: i})

        # a list of item/score tuples is returned.
        self.assertEqual(self.zs.popmin(2), [(b'i0', 0.), (b'i1', 1.)])
        self.assertEqual(self.zs.popmax(2), [(b'i9', 9.), (b'i8', 8.)])

        # when called with no args, a list is still returned.
        self.assertEqual(self.zs.popmin(), [(b'i2', 2.)])
        self.assertEqual(self.zs.popmax(), [(b'i7', 7.)])

        # blocking pop returns single item.
        self.assertEqual(self.zs.bpopmin(), (b'i3', 3.))
        self.assertEqual(self.zs.bpopmax(), (b'i6', 6.))

        # blocking-pop with timeout.
        self.assertEqual(self.zs.bpopmin(2), (b'i4', 4.))
        self.assertEqual(self.zs.bpopmax(2), (b'i5', 5.))

        # empty list is returned when zset is empty.
        self.assertEqual(self.zs.popmin(), [])
        self.assertEqual(self.zs.popmax(), [])

    def test_item_apis(self):
        self.zs['i1'] = 1
        self.zs['i0'] = 0
        self.zs['i3'] = 3
        self.zs['i2'] = 2

        self.assertEqual(self.zs[0, False], [b'i0'])
        self.assertEqual(self.zs[0, True], [(b'i0', 0)])
        self.assertEqual(self.zs[2, False], [b'i2'])
        self.assertEqual(self.zs[2, True], [(b'i2', 2)])
        self.assertEqual(self.zs[-1, True], [(b'i3', 3)])
        self.assertEqual(self.zs[9, True], [])

        self.assertEqual(self.zs[0], [b'i0'])
        self.assertEqual(self.zs[2], [b'i2'])
        self.assertEqual(self.zs[9], [])

        del self.zs['i1']
        del self.zs['i3']
        self.zs['i2'] = -2
        self.zs['i9'] = 9
        self.assertZSet([(b'i2', -2.), (b'i0', 0.), (b'i9', 9.)])

    def test_slicing(self):
        self.zs.add({'i1': 1, 'i2': 2, 'i3': 3, 'i0': 0})
        self.assertEqual(self.zs[:1, True], [(b'i0', 0)])
        self.assertEqual(self.zs[1:3, False], [b'i1', b'i2'])
        self.assertEqual(self.zs[1:-1, True], [(b'i1', 1), (b'i2', 2)])

        self.assertEqual(self.zs['i1':, False], [b'i1', b'i2', b'i3'])
        self.assertEqual(self.zs[:'i2', False], [b'i0', b'i1'])
        self.assertEqual(
            self.zs['i0':'i3', True],
            [(b'i0', 0), (b'i1', 1), (b'i2', 2)])
        self.assertRaises(KeyError, self.zs.__getitem__, (slice('i9'), False))
        self.assertEqual(self.zs[99:, False], [])

        del self.zs[:'i2']
        self.assertZSet([(b'i2', 2.), (b'i3', 3.)])
        del self.zs[1:]
        self.assertZSet([(b'i2', 2.)])

    def test_combine_store(self):
        zs2 = db.ZSet('my-zset2')
        self.zs.add({1: 1, 2: 2, 3: 3})
        zs2.add({3: 3, 4: 4, 5: 5})

        zs3 = self.zs.unionstore('my-zset3', zs2)
        self.assertEqual(
            list(zs3),
            [(b'1', 1.), (b'2', 2.), (b'4', 4.), (b'5', 5.), (b'3', 6.)])

        zs3 = self.zs.interstore('my-zset3', zs2)
        self.assertEqual(list(zs3), [(b'3', 6.)])

        self.zs |= zs2
        self.assertZSet([
            (b'1', 1.), (b'2', 2.), (b'4', 4.), (b'5', 5.), (b'3', 6.)])

        zs3 &= zs2
        self.assertEqual(list(zs3), [(b'3', 9.)])

    def test_search(self):
        self.zs.add({'foo': 1, 'bar': 2, 'baz': 1, 'nug': 3})
        self.assertEqual(
            list(self.zs.search('b*')),
            [(b'baz', 1.), (b'bar', 2.)])

    def test_sort(self):
        values = ['charlie', 3, 'zaizee', 2, 'mickey', 6, 'huey', 3]
        self.zs.add(dict(zip(values[::2], values[1::2])))
        self.assertEqual(
            self.zs.sort(),
            [b'charlie', b'huey', b'mickey', b'zaizee'])

        self.zs.sort(ordering='DESC', limit=3, store='z_dest')
        res = db.List('z_dest')
        self.assertEqual(list(res), [b'zaizee', b'mickey', b'huey'])

    def test_as_items(self):
        self.zs.add({'foo': 3, 'bar': 1, 'baz': 2})
        self.assertEqual(self.zs.as_items(True),
                         [('bar', 1.), ('baz', 2.), ('foo', 3.)])
        self.assertEqual(db.ZSet('test').as_items(), [])

    def test_from_dict(self):
        data = dict(zip('abcdefghij', [float(i) for i in range(10)]))
        zs = ZSet.from_dict(db, 'test', data)
        self.assertEqual(zs.as_items(True), sorted(data.items()))


class TestList(WalrusTestCase):
    def setUp(self):
        super(TestList, self).setUp()
        self.lst = db.List('my-list')

    def test_basic_apis(self):
        self.lst.append('i1')
        self.lst.extend(['i2', 'i3'])
        self.lst.prepend('ix')
        self.assertList(self.lst, [b'ix', b'i1', b'i2', b'i3'])

        self.lst.insert('iy', 'i2', 'before')
        self.lst.insert('iz', 'i2', 'after')
        self.assertList(self.lst, [b'ix', b'i1', b'iy', b'i2', b'iz', b'i3'])

        self.assertEqual(self.lst.pop(), b'i3')
        self.assertEqual(self.lst.popleft(), b'ix')
        self.assertEqual(len(self.lst), 4)

    def test_item_apis(self):
        self.lst.append('i0')
        self.assertEqual(self.lst[0], b'i0')

        self.lst.extend(['i1', 'i2'])
        del self.lst['i1']
        self.assertList(self.lst, [b'i0', b'i2'])

        self.lst[1] = 'i2x'
        self.assertList(self.lst, [b'i0', b'i2x'])

        del self.lst[0]
        self.assertList(self.lst, [b'i2x'])

        del self.lst[99]
        self.assertList(self.lst, [b'i2x'])

        del self.lst['ixxx']
        self.assertList(self.lst, [b'i2x'])

    def test_slicing(self):
        self.lst.extend(['i1', 'i2', 'i3', 'i4'])
        self.assertEqual(self.lst[:1], [b'i1'])
        self.assertEqual(self.lst[:2], [b'i1', b'i2'])
        self.assertEqual(self.lst[:-1], [b'i1', b'i2', b'i3'])
        self.assertEqual(self.lst[1:2], [b'i2'])
        self.assertEqual(self.lst[1:], [b'i2', b'i3', b'i4'])

        l = db.List('l1')
        l.extend(range(10))

        # LTRIM, preserve the 1st to last (removes the 0th element).
        del l[1:-1]
        self.assertEqual([int(decode(i)) for i in l],
                         [1, 2, 3, 4, 5, 6, 7, 8, 9])

        # Trim the list so that it contains only the values within the
        # specified range.
        del l[:3]
        self.assertEqual([int(decode(i)) for i in l], [1, 2, 3])

    def test_sort(self):
        values = ['charlie', 'zaizee', 'mickey', 'huey']
        self.lst.extend(values)
        self.assertEqual(self.lst.sort(),
                         [b'charlie', b'huey', b'mickey', b'zaizee'])

        self.lst.sort(ordering='DESC', limit=3, store='l_dest')
        self.assertList(db.List('l_dest'), [b'zaizee', b'mickey', b'huey'])

    def test_as_list(self):
        self.lst.extend(['foo', 'bar'])
        self.assertEqual(self.lst.as_list(True), ['foo', 'bar'])
        self.assertEqual(db.List('test').as_list(), [])

    def test_from_list(self):
        data = list('abcdefghij')
        lst = List.from_list(db, 'test', data)
        self.assertEqual(lst.as_list(True), data)


class TestArray(WalrusTestCase):
    def setUp(self):
        super(TestArray, self).setUp()
        self.arr = db.Array('my-arr')

    def test_basic_apis(self):
        self.arr.append('i1')
        self.arr.append('i2')
        self.arr.append('i3')
        self.arr.append('i4')
        self.assertEqual(len(self.arr), 4)

        # Indexing works. Invalid indices return None.
        self.assertEqual(self.arr[0], b'i1')
        self.assertEqual(self.arr[3], b'i4')
        self.assertTrue(self.arr[4] is None)

        # Negative indexing works and includes bounds-checking.
        self.assertEqual(self.arr[-1], b'i4')
        self.assertEqual(self.arr[-4], b'i1')
        self.assertTrue(self.arr[-5] is None)

        self.assertEqual(self.arr.pop(1), b'i2')
        self.assertList(self.arr, [b'i1', b'i3', b'i4'])

        self.assertEqual(self.arr.pop(), b'i4')
        self.assertList(self.arr, [b'i1', b'i3'])

        self.arr[-1] = 'iy'
        self.arr[0] = 'ix'
        self.assertList(self.arr, [b'ix', b'iy'])

        self.assertTrue('iy' in self.arr)
        self.assertFalse('i1' in self.arr)

        self.arr.extend(['foo', 'bar', 'baz'])
        self.assertList(self.arr, [b'ix', b'iy', b'foo', b'bar', b'baz'])

    def test_as_list(self):
        self.arr.extend(['foo', 'bar'])
        self.assertEqual(self.arr.as_list(True), ['foo', 'bar'])
        self.assertEqual(db.Array('test').as_list(), [])

    def test_from_list(self):
        data = list('abcdefghij')
        arr = Array.from_list(db, 'test', data)
        self.assertEqual(arr.as_list(True), data)


class TestStream(WalrusTestCase):
    def setUp(self):
        super(TestStream, self).setUp()
        db.delete('my-stream')
        db.delete('sa')
        db.delete('sb')

    def _create_test_data(self):
        return (db.xadd('sa', {'k': 'a1'}, b'1'),
                db.xadd('sb', {'k': 'b1'}, b'2'),
                db.xadd('sa', {'k': 'a2'}, b'3'),
                db.xadd('sb', {'k': 'b2'}, b'4'),
                db.xadd('sb', {'k': 'b3'}, b'5'))

    @stream_test
    def test_stream_group_info(self):
        sa = db.Stream('sa')
        ra1 = sa.add({'k': 'a1'})
        ra2 = sa.add({'k': 'a2'})
        ra3 = sa.add({'k': 'a3'})

        sb = db.Stream('sb')
        rb1 = sb.add({'k': 'b1'})

        sa_info = sa.info()
        self.assertEqual(sa_info['groups'], 0)
        self.assertEqual(sa_info['length'], 3)
        self.assertEqual(sa_info['first-entry'][0], ra1)
        self.assertEqual(sa_info['last-entry'][0], ra3)

        sb_info = sb.info()
        self.assertEqual(sb_info['groups'], 0)
        self.assertEqual(sb_info['length'], 1)
        self.assertEqual(sb_info['first-entry'][0], rb1)
        self.assertEqual(sb_info['last-entry'][0], rb1)

        self.assertEqual(sa.groups_info(), [])
        self.assertEqual(sb.groups_info(), [])

        # Create consumer groups.
        cga = db.consumer_group('cga', ['sa'])
        cga.create()
        cgab = db.consumer_group('cgab', ['sa', 'sb'])
        cgab.create()

        self.assertEqual(sa.info()['groups'], 2)
        self.assertEqual(sb.info()['groups'], 1)

        sa_groups = sa.groups_info()
        self.assertEqual(len(sa_groups), 2)
        self.assertEqual(sorted(g['name'] for g in sa_groups),
                         [b'cga', b'cgab'])

        sb_groups = sb.groups_info()
        self.assertEqual(len(sb_groups), 1)
        self.assertEqual(sb_groups[0]['name'], b'cgab')

        # Verify we can get stream info from the consumer group.
        stream_info = cgab.stream_info()
        self.assertEqual(sorted(stream_info), ['sa', 'sb'])

        # Destroy consumer group?
        cgab.destroy()
        self.assertEqual(len(sa.groups_info()), 1)
        self.assertEqual(len(sb.groups_info()), 0)

    @stream_test
    def test_consumer_group_create(self):
        cg = db.consumer_group('cg', ['sa'])
        self.assertEqual(cg.create(), {'sa': True})

        # Creating the consumer group again will report that it was not created
        # for the given key(s).
        self.assertEqual(cg.create(), {'sa': False})

        # We can register the consumer group with another key.
        cg = db.consumer_group('cg', ['sa', 'sb'])
        self.assertEqual(cg.create(), {'sa': False, 'sb': True})

    @stream_test
    def test_consumer_group_stream_creation(self):
        cg = db.consumer_group('cg1', ['stream-a', 'stream-b'])
        self.assertFalse(db.exists('stream-a'))
        self.assertFalse(db.exists('stream-b'))

        cg.create()

        # The streams were created (by adding and then deleting a message).
        self.assertTrue(db.exists('stream-a'))
        self.assertTrue(db.exists('stream-b'))

        # The streams that were automatically created will not have any data.
        self.assertEqual(db.xlen('stream-a'), 0)
        self.assertEqual(db.xlen('stream-b'), 0)

        # If a stream already exists that's OK.
        db.xadd('stream-c', {'data': 'dummy'}, id=b'1')
        cg = db.consumer_group('cg2', ['stream-c', 'stream-d'])
        self.assertTrue(db.exists('stream-c'))
        self.assertEqual(db.type('stream-c'), b'stream')
        self.assertFalse(db.exists('stream-d'))

        cg.create()
        self.assertTrue(db.exists('stream-d'))
        self.assertEqual(db.type('stream-c'), b'stream')
        self.assertEqual(db.type('stream-d'), b'stream')
        self.assertEqual(db.xlen('stream-c'), 1)
        self.assertEqual(db.xlen('stream-d'), 0)

        # If a stream key already exists and is a different type, fail.
        db.lpush('l1', 'item-1')
        db.hset('h1', 'key', 'data')
        db.sadd('s1', 'item-1')
        db.set('k1', 'v1')
        db.zadd('z1', {'item-1': 1.0})
        for key in ('l1', 'h1', 's1', 'k1', 'z1'):
            cg = db.consumer_group('cg-%s' % key, keys=[key])
            self.assertRaises(ValueError, cg.create)

    @stream_test
    def test_consumer_group_streams(self):
        ra1, rb1, ra2, rb2, rb3 = self._create_test_data()
        cg = db.consumer_group('g1', ['sa', 'sb'])

        self.assertEqual(cg.sa[ra1], (ra1, {b'k': b'a1'}))
        self.assertEqual(cg.sb[rb3], (rb3, {b'k': b'b3'}))

        def assertMessages(resp, expected):
            self.assertEqual([mid for mid, _ in resp], expected)

        assertMessages(cg.sa[ra1:], [ra1, ra2])
        assertMessages(cg.sa[:ra1], [ra1])
        assertMessages(cg.sa[ra2:], [ra2])
        assertMessages(cg.sa[:ra2], [ra1, ra2])
        assertMessages(cg.sa[rb3:], [])
        assertMessages(cg.sa[:b'0-1'], [])
        assertMessages(list(cg.sa), [ra1, ra2])

        assertMessages(cg.sb[rb1:], [rb1, rb2, rb3])
        assertMessages(cg.sb[rb1::2], [rb1, rb2])
        assertMessages(cg.sb[:rb1], [rb1])
        assertMessages(cg.sb[rb3:], [rb3])
        assertMessages(cg.sb[:rb3], [rb1, rb2, rb3])
        assertMessages(list(cg.sb), [rb1, rb2, rb3])

        self.assertEqual(len(cg.sa), 2)
        self.assertEqual(len(cg.sb), 3)

        del cg.sa[ra1]
        del cg.sb[rb1, rb3]
        self.assertEqual(len(cg.sa), 1)
        self.assertEqual(len(cg.sb), 1)
        assertMessages(list(cg.sa), [ra2])
        assertMessages(list(cg.sb), [rb2])

    @stream_test
    def test_consumer_group_container(self):
        ra1, rb1, ra2, rb2, rb3 = self._create_test_data()
        cg1 = db.consumer_group('g1', {'sa': '1', 'sb': '0'})
        cg2 = db.consumer_group('g2', {'sb': '2'})

        self.assertEqual(cg1.create(), {'sa': True, 'sb': True})
        self.assertEqual(cg2.create(), {'sb': True})

        self.assertEqual(dict(cg1.read(count=2)), {
            b'sa': [(ra2, {b'k': b'a2'})],
            b'sb': [(rb1, {b'k': b'b1'}), (rb2, {b'k': b'b2'})]})
        self.assertEqual(cg1.sa.read(), [])
        self.assertEqual(cg1.sb.read(), [(rb3, {b'k': b'b3'})])

        self.assertEqual(cg1.sa.ack(ra2), 1)
        self.assertEqual(cg1.sb.ack(rb1, rb3), 2)
        p1, = cg1.sb.pending()
        self.assertEqual(p1['message_id'], rb2)
        self.assertEqual(p1['consumer'], b'g1.c1')

        self.assertEqual(cg2.read(count=1), [
            [b'sb', [(rb2, {b'k': b'b2'})]]])
        self.assertEqual(cg2.sb.read(), [(rb3, {b'k': b'b3'})])

        self.assertEqual(cg1.destroy(), {'sa': 1, 'sb': 1})
        self.assertEqual(cg2.destroy(), {'sb': 1})

    @stream_test
    def test_consumer_group_consumers(self):
        ra1, rb1, ra2, rb2, rb3 = self._create_test_data()
        cg11 = db.consumer_group('g1', {'sa': '0', 'sb': '0'}, consumer='cg11')
        cg11.create()
        cg12 = cg11.consumer('cg12')

        self.assertEqual(dict(cg11.read(count=1)), {
            b'sa': [(ra1, {b'k': b'a1'})],
            b'sb': [(rb1, {b'k': b'b1'})]})

        self.assertEqual(dict(cg12.read(count=1, block=1)), {
            b'sa': [(ra2, {b'k': b'a2'})],
            b'sb': [(rb2, {b'k': b'b2'})]})

        pa1, pa2 = cg11.sa.pending()
        self.assertEqual(pa1['message_id'], ra1)
        self.assertEqual(pa1['consumer'], b'cg11')
        self.assertEqual(pa2['message_id'], ra2)
        self.assertEqual(pa2['consumer'], b'cg12')

        pb1, pb2 = cg11.sb.pending()
        self.assertEqual(pb1['message_id'], rb1)
        self.assertEqual(pb1['consumer'], b'cg11')
        self.assertEqual(pb2['message_id'], rb2)
        self.assertEqual(pb2['consumer'], b'cg12')

    @stream_test
    def test_read_api(self):
        sa = db.Stream('a')
        sb = db.Stream('b')
        sc = db.Stream('c')
        streams = [sa, sb, sc]
        docids = []
        for i in range(20):
            stream = streams[i % 3]
            docids.append(stream.add({'k': 'v%s' % i}, id=i + 1))

        def assertData(ret, idxs, is_multi=False):
            if is_multi:
                ret = dict(ret)
                accum = {}
                for idx in idxs:
                    sname = encode('abc'[idx % 3])
                    accum.setdefault(sname, [])
                    accum[sname].append((
                        docids[idx], {b'k': encode('v%s' % idx)}))
            else:
                accum = []
                for idx in idxs:
                    accum.append((docids[idx], {b'k': encode('v%s' % idx)}))
            self.assertEqual(ret, accum)

        assertData(sa.read(), [0, 3, 6, 9, 12, 15, 18])
        assertData(sc.read(), [2, 5, 8, 11, 14, 17])

        # We can specify a maximum number of records via "count".
        assertData(sa.read(3), [0, 3, 6])
        assertData(sb.read(2), [1, 4])
        assertData(sc.read(4), [2, 5, 8, 11])

        # We get the same values we read earlier.
        assertData(sa.read(2), [0, 3])

        # We can pass a minimum ID and will get newer data -- even if the ID
        # does not exist in the stream. We can also pass an exact ID and unlike
        # the range function, it is not inclusive.
        assertData(sa.read(2, last_id=docids[3]), [6, 9])
        assertData(sa.read(2, last_id=docids[4]), [6, 9])

        # If the last ID exceeds the highest ID (indicating no data), None is
        # returned. This is the same whether or not "count" is specified.
        self.assertEqual(sa.read(last_id=docids[18]), [])
        self.assertEqual(sa.read(2, last_id=docids[18]), [])

        # The count is a maximum, so up-to 2 items are return -- but since only
        # one item in the stream exceeds the given ID, we only get one result.
        assertData(sa.read(2, last_id=docids[17]), [18])

        # If a timeout is set and any stream can return a value, then that
        # value is returned immediately.
        assertData(sa.read(2, block=1, last_id=docids[17]), [18])
        assertData(sb.read(2, block=1, last_id=docids[18]), [19])

        # If no items are available and we timed-out, None is returned.
        self.assertEqual(sc.read(block=1, last_id=docids[19]), [])
        self.assertEqual(sc.read(2, block=1, last_id=docids[19]), [])

        # When multiple keys are given, up-to "count" items per stream
        # are returned.
        normalized = _normalize_stream_keys(['a', 'b', 'c'])
        res = db.xread(normalized, count=2)
        assertData(res, [0, 1, 2, 3, 4, 5], True)

        # Specify max-ids for each stream. The max value in "c" is 17, so
        # nothing will be returned for "c".
        uids = [decode(docid) for docid in docids]
        res = db.xread({'a': uids[15], 'b': uids[16], 'c': uids[17]},
                       count=3)
        assertData(res, [18, 19], True)

        # Now we limit ourselves to being able to pull only a single item from
        # stream "c".
        res = db.xread({'a': uids[18], 'b': uids[19], 'c': uids[16]})
        assertData(res, [17], True)

        # None is returned when no results are present and timeout is None or
        # if we reach the timeout.
        res = db.xread({'a': uids[18], 'b': uids[19], 'c': uids[17]})
        self.assertEqual(res, [])

        res = db.xread({'a': uids[18], 'b': uids[19], 'c': uids[17]},
                       count=1, block=1)
        self.assertEqual(res, [])

    @stream_test
    def test_set_id_stream(self):
        stream = db.Stream('my-stream')
        stream.add({'k': 'v1'}, id='3')
        self.assertTrue(stream.set_id('5'))
        self.assertRaises(Exception, stream.add, {'k': 'v2'}, id='4')
        stream.add({'k': 'v3'}, id='6')
        self.assertEqual(stream.read(), [
            (b'3-0', {b'k': b'v1'}),
            (b'6-0', {b'k': b'v3'})])

    @stream_test
    def test_basic_apis(self):
        stream = db.Stream('my-stream')

        # Item ids will be 1-0, 11-0, ...91-0.
        item_ids = [stream.add({'k': 'v%s' % i}, id='%s1' % i)
                    for i in range(10)]
        self.assertEqual(len(stream), 10)

        # Redis automatically adds the sequence number.
        self.assertEqual(item_ids[:3], [b'1-0', b'11-0', b'21-0'])
        self.assertEqual(item_ids[7:], [b'71-0', b'81-0', b'91-0'])

        def assertData(items, expected):
            self.assertEqual(items, [(item_ids[e], {b'k': encode('v%s' % e)})
                                     for e in expected])

        # The sequence number is optional if it's zero.
        assertData(stream[:'1'], [0])
        assertData(stream[:'1-0'], [0])
        assertData(stream['91':], [9])
        assertData(stream['91-0':], [9])
        assertData(stream['91-1':], [])

        # We can slice up to a value. If the sequence number is omitted it will
        # be treated as zero.
        assertData(stream[:'31'], [0, 1, 2, 3])
        assertData(stream[:'31-0'], [0, 1, 2, 3])
        assertData(stream[:'31-1'], [0, 1, 2, 3])

        # We can slice up from a value as well.
        assertData(stream['71':], [7, 8, 9])
        assertData(stream['71-0':], [7, 8, 9])
        assertData(stream['71-1':], [8, 9])

        # We can also slice between values.
        assertData(stream['21':'41'], [2, 3, 4])
        assertData(stream['21-0':'41'], [2, 3, 4])
        assertData(stream['21':'41-0'], [2, 3, 4])
        assertData(stream['21-1':'41'], [3, 4])
        assertData(stream['21-1':'41-1'], [3, 4])

        # The "step" parameter, the third part of the slice, indicates count.
        assertData(stream['41'::3], [4, 5, 6])
        assertData(stream[:'41':3], [0, 1, 2])
        assertData(stream['81'::3], [8, 9])

        # Test using in-between values. The endpoints of the slice are
        # inclusive.
        assertData(stream[:'5'], [0])
        assertData(stream[:'5-1'], [0])
        assertData(stream[:'25'], [0, 1, 2])
        assertData(stream[:'25-1'], [0, 1, 2])
        assertData(stream['25':'55'], [3, 4, 5])
        assertData(stream['55':'92'], [6, 7, 8, 9])
        assertData(stream['91':'92'], [9])

        # If we go above or below, it returns an empty list.
        assertData(stream['92':], [])
        assertData(stream[:'0'], [])

        # We can also provide a count when indexing in-between.
        assertData(stream['25':'55':2], [3, 4])
        assertData(stream['55':'92':1], [6])

        # Use "del" to remove items by ID. The sequence number will be treated
        # as zero if not provided.
        del stream['21', '41-0', '61']
        del stream['51-1']  # Has no effect since we only have 51-0.
        assertData(stream['5':'65'], [1, 3, 5])
        self.assertEqual(len(stream), 7)

        del stream['21']  # Can delete non-existent items.

        # Cannot add lower than maximum ID.
        self.assertRaises(Exception, stream.add, {'k': 'v2'}, id='90-1')
        self.assertRaises(Exception, stream.add, {'k': 'v2'}, id='91-0')

        # Adding a "1" to the sequence works:
        new_id = stream.add({'k': 'v10'}, id='91-1')
        self.assertEqual(new_id, b'91-1')

        # Length reflects the latest addition.
        self.assertEqual(len(stream), 8)

        # Range starting at 91-0 yields 91-0 and 91-1.
        data = stream['91-0':]
        self.assertEqual(len(data), 2)
        self.assertEqual([obj_id for obj_id, _ in data], [b'91-0', b'91-1'])

        # Remove the two 91-x items.
        del stream['91', '91-1']

        # Sanity check that the data was really remove.
        self.assertEqual(len(stream), 6)
        assertData(stream['61':], [7, 8])

        # Can we add an item with an id lower than 91? We've deleted it so the
        # last value is 81, but this still doesn't work (?).
        for docid in ('90', '91', '91-1'):
            self.assertRaises(Exception, stream.add, {'k': 'v9'}, id='90')

        new_id = stream.add({'k': 'v9'}, id='91-2')
        self.assertEqual(new_id, b'91-2')
        self.assertEqual(stream['91':], [(b'91-2', {b'k': b'v9'})])
        del stream['91-2']

        nremoved = stream.trim(4, approximate=False)
        self.assertEqual(nremoved, 2)
        assertData(stream[:], [3, 5, 7, 8])

        # Trimming again returns 0, no items removed.
        self.assertEqual(stream.trim(4, approximate=False), 0)

        # Verify we can iterate over the stream.
        assertData(list(stream), [3, 5, 7, 8])

        # Verify we can get items by id.
        d5 = stream.get('51-0')
        self.assertEqual(d5, (b'51-0', {b'k': b'v5'}))

        # Nonexistant values return None.
        self.assertTrue(stream.get('61-0') is None)


class TestBitField(WalrusTestCase):
    def setUp(self):
        super(TestBitField, self).setUp()
        self.bf = db.bit_field('bf')

    def test_simple_operations(self):
        resp = (self.bf
                .set('u8', 8, 255)
                .get('u8', 0)
                .get('u4', 8)  # 1111
                .get('u4', 12)  # 1111
                .get('u4', 13)  # 1110
                .execute())
        self.assertEqual(resp, [0, 0, 15, 15, 14])

        resp = (self.bf
                .set('u8', 4, 1)  # 00ff -> 001f (returns old val, 0x0f)
                .get('u16', 0)  # 001f (00011111)
                .set('u16', 0, 0))  # 001f -> 0000
        self.assertEqual(list(resp), [15, 31, 31])

        resp = (self.bf
                .incrby('u8', 8, 254)
                .get('u16', 0))
        self.assertEqual(list(resp), [254, 254])

        # Verify overflow protection works:
        resp = (self.bf
                .incrby('u8', 8, 2, 'FAIL')
                .incrby('u8', 8, 1)
                .incrby('u8', 8, 1)  # Still "FAIL".
                .get('u16', 0))
        self.assertEqual(list(resp), [None, 255, None, 255])

        self.assertEqual(self.bf.get_raw(), b'\x00\xff')

    def test_slicing(self):
        resp = self.bf.set('u8', 0, 166).execute()  # 10100110

        self.assertEqual(self.bf[:8], 166)
        self.assertEqual(self.bf[:4], 10)  # 1010
        self.assertEqual(self.bf[4:8], 6)   # 0110
        self.assertEqual(self.bf[2:6], 9) # 1001
        self.assertEqual(self.bf[6:10], 8) # 10?? -> 1000
        self.assertEqual(self.bf[8:16], 0)  # Undefined, defaults to zero.

        self.assertRaises(ValueError, lambda: self.bf[1])
        self.assertRaises(ValueError, lambda: self.bf[1:])
        self.assertRaises(ValueError, lambda: self.bf[4:1])

        self.bf[:8] = 89  # 01011001
        self.assertEqual(self.bf[:8], 89)
        def overflow():
            self.bf[:8] = 256
        self.assertRaises(ValueError, overflow)
        self.bf[:8] = 255
        self.assertEqual(self.bf[:8], 255)

        del self.bf[2:6]
        self.assertEqual(self.bf[:8], 195)  # 11000011


class TestBloomFilter(WalrusTestCase):
    def setUp(self):
        super(TestBloomFilter, self).setUp()
        self.bf = db.bloom_filter('bf')

    def test_bloom_filter(self):
        data = ('foo', 'bar', 'baz', 'nugget', 'this is a test', 'testing',
                'alpha', 'beta', 'delta', 'gamma')

        # Verify that the bloom-filter does not contain any of our items.
        for item in data:
            self.assertFalse(item in self.bf)

        # Add all the items to the bloom filter.
        for item in data:
            self.bf.add(item)

        # Verify that all of our items are now present.
        for item in data:
            self.assertTrue(item in self.bf)

        # Making some small modifications we can verify that all these other
        # items are not present, however.
        for item in data:
            self.assertFalse(item.upper() in self.bf)
            self.assertFalse(item.title() in self.bf)



from functools import wraps
import glob
import os
import sys
import threading
import uuid

try:
    from redis import Redis
    from redis.exceptions import ConnectionError
    from redis.exceptions import TimeoutError
except ImportError:
    Redis = object
    ConnectionError = TimeoutError = Exception

from walrus.autocomplete import Autocomplete
from walrus.cache import Cache
from walrus.containers import Array
from walrus.containers import BitField
from walrus.containers import BloomFilter
from walrus.containers import ConsumerGroup
from walrus.containers import Hash
from walrus.containers import HyperLogLog
from walrus.containers import List
from walrus.containers import Set
from walrus.containers import Stream
from walrus.containers import ZSet
from walrus.counter import Counter
from walrus.fts import Index
from walrus.graph import Graph
from walrus.lock import Lock
from walrus.rate_limit import RateLimit
from walrus.rate_limit import RateLimitLua
from walrus.streams import TimeSeries


class TransactionLocal(threading.local):
    def __init__(self, **kwargs):
        super(TransactionLocal, self).__init__(**kwargs)
        self.pipes = []

    @property
    def pipe(self):
        if len(self.pipes):
            return self.pipes[-1]

    def commit(self):
        pipe = self.pipes.pop()
        return pipe.execute()

    def abort(self):
        pipe = self.pipes.pop()
        pipe.reset()


class Database(Redis):
    """
    Redis-py client with some extras.
    """
    def __init__(self, *args, **kwargs):
        """
        :param args: Arbitrary positional arguments to pass to the
            base ``Redis`` instance.
        :param kwargs: Arbitrary keyword arguments to pass to the
            base ``Redis`` instance.
        :param str script_dir: Path to directory containing walrus
            scripts. Use "script_dir=False" to disable loading any scripts.
        """
        script_dir = kwargs.pop('script_dir', None)
        super(Database, self).__init__(*args, **kwargs)
        self.__mapping = {
            b'list': self.List,
            b'set': self.Set,
            b'zset': self.ZSet,
            b'hash': self.Hash}
        self._transaction_local = TransactionLocal()
        self._transaction_lock = threading.RLock()
        if script_dir is not False:
            self.init_scripts(script_dir=script_dir)

    def __bool__(self):
        return True  # Avoid falling back to __len__().

    def xsetid(self, name, id):
        """
        Set the last ID of the given stream.

        :param name: stream identifier
        :param id: new value for last ID
        """
        return self.execute_command('XSETID', name, id) == b'OK'

    def xpending_summary(self, key, group):
        """
        Pending message summary report.

        :param key: stream identifier
        :param group: consumer group name
        :returns: dictionary of information about pending messages
        """
        return self.xpending(key, group)

    def get_transaction(self):
        with self._transaction_lock:
            local = self._transaction_local
            local.pipes.append(self.pipeline())
            return local.pipe

    def commit_transaction(self):
        """
        Commit the currently active transaction (Pipeline). If no
        transaction is active in the current thread, an exception
        will be raised.

        :returns: The return value of executing the Pipeline.
        :raises: ``ValueError`` if no transaction is active.
        """
        with self._transaction_lock:
            local = self._transaction_local
            if not local.pipes:
                raise ValueError('No transaction is currently active.')
            return local.commit()

    def clear_transaction(self):
        """
        Clear the currently active transaction (if exists). If the
        transaction stack is not empty, then a new pipeline will
        be initialized.

        :returns: No return value.
        :raises: ``ValueError`` if no transaction is active.
        """
        with self._transaction_lock:
            local = self._transaction_local
            if not local.pipes:
                raise ValueError('No transaction is currently active.')
            local.abort()

    def atomic(self):
        return _Atomic(self)

    def init_scripts(self, script_dir=None):
        self._scripts = {}
        if not script_dir:
            script_dir = os.path.join(os.path.dirname(__file__), 'scripts')
        for filename in glob.glob(os.path.join(script_dir, '*.lua')):
            with open(filename, 'r') as fh:
                script_obj = self.register_script(fh.read())
                script_name = os.path.splitext(os.path.basename(filename))[0]
                self._scripts[script_name] = script_obj

    def run_script(self, script_name, keys=None, args=None):
        """
        Execute a walrus script with the given arguments.

        :param script_name: The base name of the script to execute.
        :param list keys: Keys referenced by the script.
        :param list args: Arguments passed in to the script.
        :returns: Return value of script.

        .. note:: Redis scripts require two parameters, ``keys``
            and ``args``, which are referenced in lua as ``KEYS``
            and ``ARGV``.
        """
        return self._scripts[script_name](keys, args)

    def get_temp_key(self):
        """
        Generate a temporary random key using UUID4.
        """
        return 'temp.%s' % uuid.uuid4()

    def __iter__(self):
        """
        Iterate over the keys of the selected database.
        """
        return iter(self.scan_iter())

    def __len__(self):
        return self.dbsize()

    def search(self, pattern):
        """
        Search the keyspace of the selected database using the
        given search pattern.

        :param str pattern: Search pattern using wildcards.
        :returns: Iterator that yields matching keys.
        """
        return self.scan_iter(pattern)

    def get_key(self, key):
        """
        Return a rich object for the given key. For instance, if
        a hash key is requested, then a :py:class:`Hash` will be
        returned.

        Note: only works for Hash, List, Set and ZSet.

        :param str key: Key to retrieve.
        :returns: A hash, set, list, zset or array.
        """
        return self.__mapping.get(self.type(key), self.__getitem__)(key)

    def hash_exists(self, key):
        return self.exists(key)

    def autocomplete(self, namespace='autocomplete', **kwargs):
        return Autocomplete(self, namespace, **kwargs)

    def cache(self, name='cache', default_timeout=3600):
        """
        Create a :py:class:`Cache` instance.

        :param str name: The name used to prefix keys used to
            store cached data.
        :param int default_timeout: The default key expiry.
        :returns: A :py:class:`Cache` instance.
        """
        return Cache(self, name=name, default_timeout=default_timeout)

    def counter(self, name):
        """
        Create a :py:class:`Counter` instance.

        :param str name: The name used to store the counter's value.
        :returns: A :py:class:`Counter` instance.
        """
        return Counter(self, name=name)

    def graph(self, name, *args, **kwargs):
        """
        Creates a :py:class:`Graph` instance.

        :param str name: The namespace for the graph metadata.
        :returns: a :py:class:`Graph` instance.
        """
        return Graph(self, name, *args, **kwargs)

    def lock(self, name, ttl=None, lock_id=None):
        """
        Create a named :py:class:`Lock` instance. The lock implements
        an API similar to the standard library's ``threading.Lock``,
        and can also be used as a context manager or decorator.

        :param str name: The name of the lock.
        :param int ttl: The time-to-live for the lock in milliseconds
            (optional). If the ttl is ``None`` then the lock will not
            expire.
        :param str lock_id: Optional identifier for the lock instance.
        """
        return Lock(self, name, ttl, lock_id)

    def rate_limit(self, name, limit=5, per=60, debug=False):
        """
        Rate limit implementation. Allows up to `limit` of events every `per`
        seconds.

        See :ref:`rate-limit` for more information.
        """
        return RateLimit(self, name, limit, per, debug)

    def rate_limit_lua(self, name, limit=5, per=60, debug=False):
        """
        Rate limit implementation. Allows up to `limit` of events every `per`
        seconds. Uses a Lua script for atomicity.

        See :ref:`rate-limit` for more information.
        """
        return RateLimitLua(self, name, limit, per, debug)

    def Index(self, name, **options):
        """
        Create a :py:class:`Index` full-text search index with the given
        name and options.
        """
        return Index(self, name, **options)

    def List(self, key):
        """
        Create a :py:class:`List` instance wrapping the given key.
        """
        return List(self, key)

    def Hash(self, key):
        """
        Create a :py:class:`Hash` instance wrapping the given key.
        """
        return Hash(self, key)

    def Set(self, key):
        """
        Create a :py:class:`Set` instance wrapping the given key.
        """
        return Set(self, key)

    def ZSet(self, key):
        """
        Create a :py:class:`ZSet` instance wrapping the given key.
        """
        return ZSet(self, key)

    def HyperLogLog(self, key):
        """
        Create a :py:class:`HyperLogLog` instance wrapping the given
        key.
        """
        return HyperLogLog(self, key)

    def Array(self, key):
        """
        Create a :py:class:`Array` instance wrapping the given key.
        """
        return Array(self, key)

    def Stream(self, key):
        """
        Create a :py:class:`Stream` instance wrapping the given key.
        """
        return Stream(self, key)

    def consumer_group(self, group, keys, consumer=None):
        """
        Create a named :py:class:`ConsumerGroup` instance for the given key(s).

        :param group: name of consumer group
        :param keys: stream identifier(s) to monitor. May be a single stream
            key, a list of stream keys, or a key-to-minimum id mapping. The
            minimum id for each stream should be considered an exclusive
            lower-bound. The '$' value can also be used to only read values
            added *after* our command started blocking.
        :param consumer: name for consumer within group
        :returns: a :py:class:`ConsumerGroup` instance
        """
        return ConsumerGroup(self, group, keys, consumer=consumer)

    def time_series(self, group, keys, consumer=None):
        """
        Create a named :py:class:`TimeSeries` consumer-group for the
        given key(s). TimeSeries objects are almost identical to
        :py:class:`ConsumerGroup` except they offer a higher level of
        abstraction and read/write message ids as datetimes.

        :param group: name of consumer group
        :param keys: stream identifier(s) to monitor. May be a single stream
            key, a list of stream keys, or a key-to-minimum id mapping. The
            minimum id for each stream should be considered an exclusive
            lower-bound. The '$' value can also be used to only read values
            added *after* our command started blocking.
        :param consumer: name for consumer within group
        :returns: a :py:class:`TimeSeries` instance
        """
        return TimeSeries(self, group, keys, consumer=consumer)

    def bit_field(self, key):
        """
        Container for working with the Redis BITFIELD command.

        :returns: a :py:class:`BitField` instance.
        """
        return BitField(self, key)

    def bloom_filter(self, key, size=64 * 1024):
        """
        Create a :py:class:`BloomFilter` container type.

        Bloom-filters are probabilistic data-structures that are used to answer
        the question: "is X a member of set S?" It is possible to receive a
        false positive, but impossible to receive a false negative (in other
        words, if the bloom filter contains a value, it will never erroneously
        report that it does *not* contain such a value). The accuracy of the
        bloom-filter and the likelihood of a false positive can be reduced by
        increasing the size of the bloomfilter. The default size is 64KB (or
        524,288 bits).
        """
        return BloomFilter(self, key, size)

    def cas(self, key, value, new_value):
        """
        Perform an atomic compare-and-set on the value in "key", using a prefix
        match on the provided value.
        """
        return self.run_script('cas', keys=[key], args=[value, new_value])

    def listener(self, channels=None, patterns=None, is_async=False):
        """
        Decorator for wrapping functions used to listen for Redis
        pub-sub messages.

        The listener will listen until the decorated function
        raises a ``StopIteration`` exception.

        :param list channels: Channels to listen on.
        :param list patterns: Patterns to match.
        :param bool is_async: Whether to start the listener in a
            separate thread.
        """
        def decorator(fn):
            _channels = channels or []
            _patterns = patterns or []

            @wraps(fn)
            def inner():
                pubsub = self.pubsub()

                def listen():
                    for channel in _channels:
                        pubsub.subscribe(channel)
                    for pattern in _patterns:
                        pubsub.psubscribe(pattern)

                    for data_dict in pubsub.listen():
                        try:
                            ret = fn(**data_dict)
                        except StopIteration:
                            pubsub.close()
                            break

                if is_async:
                    worker = threading.Thread(target=listen)
                    worker.start()
                    return worker
                else:
                    listen()

            return inner
        return decorator

    def stream_log(self, callback, connection_id='monitor'):
        """
        Stream Redis activity one line at a time to the given
        callback.

        :param callback: A function that accepts a single argument,
            the Redis command.
        """
        conn = self.connection_pool.get_connection(connection_id, None)
        conn.send_command('monitor')
        while callback(conn.read_response()):
            pass


class _Atomic(object):
    def __init__(self, db):
        self.db = db

    @property
    def pipe(self):
        return self.db._transaction_local.pipe

    def __enter__(self):
        self.db.get_transaction()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type:
            self.clear(False)
        else:
            self.commit(False)

    def commit(self, begin_new=True):
        ret = self.db.commit_transaction()
        if begin_new:
            self.db.get_transaction()
        return ret

    def clear(self, begin_new=True):
        self.db.clear_transaction()
        if begin_new:
            self.db.get_transaction()



import datetime

from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


cache = db.cache(name='test.cache')


@cache.cached(timeout=60)
def now(seed=None):
    return datetime.datetime.now()


class Clock(object):
    @cache.cached_property()
    def now(self):
        return datetime.datetime.now()


class TestCache(WalrusTestCase):
    def test_cache_apis(self):
        # Nonexistant key returns None.
        self.assertTrue(cache.get('foo') is None)

        # Set key, value and expiration in seconds.
        self.assertEqual(cache.set('foo', 'bar', 60), True)
        self.assertEqual(cache.get('foo'), 'bar')

        self.assertEqual(cache.delete('foo'), 1)
        self.assertTrue(cache.get('foo') is None)
        self.assertEqual(cache.delete('foo'), 0)

    def test_cache_bulk_apis(self):
        self.assertEqual(cache.get_many(['k1', 'k2']), {})

        data = {'k1': 'v1', 'k2': 'v2'}
        self.assertEqual(cache.set_many(data, 60), True)
        self.assertEqual(cache.get_many(['k1', 'kx', 'k2']), data)
        self.assertEqual(cache.delete_many(['k1', 'kx', 'k2']), 2)
        self.assertEqual(cache.get_many(['k1', 'k2']), {})
        self.assertEqual(cache.delete_many(['k1', 'kx', 'k2']), 0)

    def test_cache_decorator(self):
        n0 = now()  # Each should have its own cache-key.
        n1 = now(1)
        n2 = now(2)

        self.assertTrue(n0 != n1 != n2)
        self.assertEqual(now(), n0)
        self.assertEqual(now(1), n1)
        self.assertEqual(now(2), n2)

        now.bust(1)
        self.assertNotEqual(now(1), n1)
        self.assertEqual(now(1), now(1))

    def test_cached_property(self):
        c = Clock()
        n1 = c.now
        n2 = c.now
        self.assertEqual(n1, n2)

        del c.now  # Property deleter busts the cache.
        n3 = c.now
        self.assertTrue(n1 != n3)
        self.assertEqual(c.now, n3)

    def test_cached_return_none(self):
        S = {'count': 0}
        @cache.cached()
        def returns_none(arg):
            S['count'] += 1

        def assertMisses(arg, n):
            returns_none(arg)
            self.assertEqual(S['count'], n)
        for i in range(3):
            assertMisses('foo', 1)
        assertMisses('bar', 2)
        assertMisses('foo', 2)

    def test_cached_async(self):
        @cache.cache_async()
        def double_value(value):
            return value * 2

        res = double_value(3)
        self.assertEqual(res(), 6)
        self.assertEqual(res(), 6)

        self.assertEqual(double_value(3)(), 6)
        self.assertEqual(double_value(4)(), 8)

    def test_flush_empty_cache(self):
        cache.set('foo', 'bar', 10)
        self.assertList(cache.keys(), ['test.cache:foo'])
        cache.flush()
        self.assertList(cache.keys(), [])



import random

from walrus.tests.base import WalrusTestCase
from walrus.tests.base import db


class TestAutocomplete(WalrusTestCase):
    test_data = (
        (1, 'testing python'),
        (2, 'testing python code'),
        (3, 'web testing python code'),
        (4, 'unit tests with python'))

    def setUp(self):
        super(TestAutocomplete, self).setUp()
        self.ac = db.autocomplete()

    def store_test_data(self, id_to_store=None):
        for obj_id, title in self.test_data:
            if id_to_store is None or obj_id == id_to_store:
                self.ac.store(obj_id, title, {
                    'obj_id': obj_id,
                    'title': title,
                    'value': obj_id % 2 == 0 and 'even' or 'odd'})

    def sort_results(self, results):
        return sorted(results, key=lambda item: item['obj_id'])

    def assertResults(self, results, expected):
        self.assertEqual([result['obj_id'] for result in results], expected)

    def test_search(self):
        self.store_test_data()

        results = self.ac.search('testing python')
        self.assertList(results, [
            {'obj_id': 1, 'title': 'testing python', 'value': 'odd'},
            {'obj_id': 2, 'title': 'testing python code', 'value': 'even'},
            {'obj_id': 3, 'title': 'web testing python code', 'value': 'odd'},
        ])

        results = self.ac.search('test')
        self.assertResults(results, [1, 2, 4, 3])

        results = self.ac.search('uni')
        self.assertResults(results, [4])

        self.assertList(self.ac.search(''), [])
        self.assertList(self.ac.search('missing'), [])
        self.assertList(self.ac.search('with'), [])

    def test_boosting(self):
        letters = ('alpha', 'beta', 'gamma', 'delta')
        n = len(letters)
        test_data = []
        for i in range(n * 3):
            obj_id = i + 1
            obj_type = 't%d' % ((i / n) + 1)
            title = 'test %s' % letters[i % n]
            self.ac.store(
                obj_id,
                title,
                {'obj_id': obj_id, 'title': title},
                obj_type)

        def assertBoosts(query, boosts, expected):
            results = self.ac.search(query, boosts=boosts)
            self.assertResults(results, expected)

        assertBoosts('alp', None, [1, 5, 9])
        assertBoosts('alp', {'t2': 1.1}, [5, 1, 9])
        assertBoosts('test', {'t3': 1.5, 't2': 1.1}, [
            9, 10, 12, 11, 5, 6, 8, 7, 1, 2, 4, 3])
        assertBoosts('alp', {'t1': 0.5}, [5, 9, 1])
        assertBoosts('alp', {'t1': 1.5, 't3': 1.6}, [9, 1, 5])
        assertBoosts('alp', {'t3': 1.5, '5': 1.6}, [5, 9, 1])

    def test_stored_boosts(self):
        id_to_type = {
            'aaa': 1,
            'aab': 2,
            'aac': 3,
            'aaab': 4,
            'bbbb': 4}
        for obj_id, obj_type in id_to_type.items():
            self.ac.store(obj_id, obj_type=obj_type)

        self.assertList(self.ac.search('aa'), ['aaa', 'aaab', 'aab', 'aac'])

        self.ac.boost_object(obj_type=2, multiplier=2)
        self.assertList(self.ac.search('aa'), ['aab', 'aaa', 'aaab', 'aac'])

        self.ac.boost_object('aac', multiplier=3)
        self.assertList(self.ac.search('aa'), ['aac', 'aab', 'aaa', 'aaab'])

        results = self.ac.search('aa', boosts={'aac': 1.5})
        self.assertList(results, ['aab', 'aac', 'aaa', 'aaab'])

    def test_limit(self):
        self.store_test_data()
        results = self.ac.search('testing', limit=1)
        self.assertResults(results, [1])

        results = self.ac.search('testing', limit=2)
        self.assertResults(results, [1, 2])

    def test_search_empty(self):
        self.assertList(self.ac.search(''), [])

    def test_chunked(self):
        for i in range(25):
            self.ac.store('foo %s' % (chr(i + ord('a')) * 2))

        ge = self.ac.search('foo', limit=21, chunk_size=5)
        results = list(ge)
        self.assertEqual(len(results), 21)
        self.assertEqual(results[0], 'foo aa')
        self.assertEqual(results[-1], 'foo uu')

    def test_scoring_proximity_to_front(self):
        self.ac.store('aa bb cc')
        self.ac.store('tt cc')

        self.assertList(self.ac.search('cc'), ['tt cc', 'aa bb cc'])

        self.ac.store('aa b cc')
        self.assertList(self.ac.search('cc'), ['tt cc', 'aa b cc', 'aa bb cc'])

    def test_simple(self):
        for _, title in self.test_data:
            self.ac.store(title)

        self.assertList(self.ac.search('testing'), [
            'testing python',
            'testing python code',
            'web testing python code'])
        self.assertList(self.ac.search('code'), [
            'testing python code',
            'web testing python code'])

        self.ac.store('z python code')
        self.assertList(self.ac.search('cod'), [
            'testing python code',
            'z python code',
            'web testing python code'])

    def test_sorting(self):
        strings = []
        for i in range(26):
            strings.append('aaaa%s' % chr(i + ord('a')))
            if i > 0:
                strings.append('aaa%sa' % chr(i + ord('a')))

        random.shuffle(strings)
        for s in strings:
            self.ac.store(s)

        self.assertList(self.ac.search('aaa'), sorted(strings))
        self.assertList(self.ac.search('aaa', limit=30), sorted(strings)[:30])

    def test_removing_objects(self):
        self.store_test_data()
        self.ac.remove(1)

        self.assertResults(self.ac.search('testing'), [2, 3])

        # Restore item 1 and remove item 2.
        self.store_test_data(1)
        self.ac.remove(2)

        self.assertResults(self.ac.search('testing'), [1, 3])

        # Item with obj_id=2 has already been removed.
        with self.assertRaises(KeyError):
            self.ac.remove(2)

    def test_tokenize_title(self):
        self.assertEqual(
            self.ac.tokenize_title('abc def ghi'),
            ['abc', 'def', 'ghi'])

        # Stop-words are removed automatically.
        self.assertEqual(self.ac.tokenize_title('a A tHe an a'), [])

        # Empty string yields an empty list.
        self.assertEqual(self.ac.tokenize_title(''), [])

        # Stop-words, punctuation, capitalization, etc.
        self.assertEqual(self.ac.tokenize_title(
            'The Best of times, the blurst of times'),
            ['times', 'blurst', 'times'])

    def test_exists(self):
        self.assertFalse(self.ac.exists('test'))
        self.ac.store('test')
        self.assertTrue(self.ac.exists('test'))

    def test_key_leaks(self):
        initial_key_count = len(db.keys())

        # Store a single item.
        self.store_test_data(1)

        # See how many keys we have in the db - check again in a bit.
        key_len = len(db.keys())

        # Store a second item.
        self.store_test_data(2)
        key_len2 = len(db.keys())

        self.assertTrue(key_len != key_len2)
        self.ac.remove(2)

        # Back to the original amount of keys we had after one item.
        self.assertEqual(len(db.keys()), key_len)

        # Remove the first item, back to original count at start.
        self.ac.remove(1)
        self.assertEqual(len(db.keys()), initial_key_count)

    def test_updating(self):
        # store(obj_id, title=None, data=None, obj_type=None).
        self.ac.store('id1', 'title baze', 'd1', 't1')
        self.ac.store('id2', 'title nugget', 'd2', 't2')
        self.ac.store('id3', 'title foo', 'd3', 't3')

        self.assertList(self.ac.search('tit'), ['d1', 'd3', 'd2'])

        # Overwrite the data for id1.
        self.ac.store('id1', 'title foo', 'D1', 't1')
        self.assertList(self.ac.search('tit'), ['D1', 'd3', 'd2'])

        # Overwrite the data with a new title, will remove the title one refs.
        self.ac.store('id1', 'Herple', 'done', 't1')
        self.assertList(self.ac.search('tit'), ['d3', 'd2'])
        self.assertList(self.ac.search('herp'), ['done'])

        # Overwrite again, capitalizing the data and changing the title.
        self.ac.store('id1', 'title baze', 'Done', 't1')
        self.assertList(self.ac.search('tit'), ['Done', 'd3', 'd2'])

        # Verify that we clean up any crap when updating.
        self.assertList(self.ac.search('herp'), [])

    def test_word_position_ordering(self):
        self.ac.store('aaaa bbbb')
        self.ac.store('bbbb cccc')
        self.ac.store('bbbb aaaa')
        self.ac.store('aaaa bbbb')

        results = self.ac.search('bb')
        self.assertList(results, ['bbbb aaaa', 'bbbb cccc', 'aaaa bbbb'])
        self.assertList(self.ac.search('aa'), ['aaaa bbbb', 'bbbb aaaa'])

        self.ac.store('aabb bbbb')
        self.assertList(self.ac.search('bb'), [
            'bbbb aaaa',
            'bbbb cccc',
            'aaaa bbbb',
            'aabb bbbb'])
        self.assertList(self.ac.search('aa'), [
            'aaaa bbbb',
            'aabb bbbb',
            'bbbb aaaa'])

        # Verify issue 9 is fixed.
        self.ac.store('foo one')
        self.ac.store('bar foo one')
        self.assertList(self.ac.search('foo'), ['foo one', 'bar foo one'])

    def test_return_all_results(self):
        phrases = ('aa bb', 'aa cc', 'bb aa cc', 'bb cc', 'cc aa bb')
        for phrase in phrases:
            self.ac.store(phrase)

        self.assertList(sorted(self.ac.list_data()), list(phrases))
        self.assertEqual(sorted(self.ac.list_titles()), list(phrases))

    def test_multiword_phrases(self):
        self.ac.store('p1', 'alpha beta gamma delta')
        self.ac.store('p2', 'beta delta zeta')
        self.ac.store('p3', 'gamma zeta iota')

        self.assertList(self.ac.search('ga del'), ['alpha beta gamma delta'])
        self.assertList(self.ac.search('be de'), [
            'beta delta zeta',
            'alpha beta gamma delta'])
        self.assertList(self.ac.search('de be'), [
            'beta delta zeta',
            'alpha beta gamma delta'])
        self.assertList(self.ac.search('bet delt'), [
            'beta delta zeta',
            'alpha beta gamma delta'])
        self.assertList(self.ac.search('delt bet'), [
            'beta delta zeta',
            'alpha beta gamma delta'])

        self.assertList(self.ac.search('delt bet alpha'),
                        ['alpha beta gamma delta'])

    def test_multiword_stopword_handling(self):
        self.ac.store('p1', 'alpha beta delta')
        self.ac.store('p2', 'alpha delta gamma')
        self.ac.store('p3', 'beta gamma')

        self.assertList(self.ac.search('a'), [
            'alpha beta delta',
            'alpha delta gamma'])
        self.assertList(self.ac.search('be'), [
            'beta gamma',
            'alpha beta delta'])
        # Here since "a" is a stopword and is not the last token, we strip it.
        self.assertList(self.ac.search('a be'), [
            'beta gamma',
            'alpha beta delta'])
        # a & be are stripped, since they are stopwords and not last token.
        self.assertList(self.ac.search('a be de'), [
            'alpha delta gamma',
            'alpha beta delta'])

        self.assertList(self.ac.search('al bet de'), [
            'alpha beta delta'])
